"""
All linkers inherit from `BaseLinker`.

Linkers provide the interface to add/update/delete links between entities.
Only certain entities can link to certain other entities. When using this
library, your tooling should show you which link methods are available
to the Entity subclass you are using.

Note that some links require additional data in the link. You must ensure
the destination object has this data before adding or updating the link
(example below). See the API documentation for exact link details.
Missing data will raise a `projectal.errors.ProjectalException` with
information about what is missing.


An instance of an entity class that inherits from a linker is able to link
to an instance of the target entity class directly.

```
# Get task and staff
task = projectal.Task.get('1b21e445-f29a...')
staff = projectal.Staff.get('1b21e445-f29a...')

# Task-to-Staff links require a 'resourceLink'
staff['resourceLink'] = {'utilization': 0.6}

# Task inherits from StaffLinker, so staff linking available
task.link_staff(staff)
```

"""

from projectal import api


class BaseLinker:
    """
    This base class provides methods for linking, updating, and deleting
    links between entities.

    Used internally by the library.
    """

    # TODO: The linker currently only lets you link 1-to-1. In the future,
    # we should detect and support linking in bulk.

    @classmethod
    def _add_link(cls, from_uuId, to_entity_name, to_entity_list, list_name=None):
        return cls._link(from_uuId, to_entity_name, to_entity_list, 'add', list_name=list_name)

    @classmethod
    def _update_link(cls, from_uuId, to_entity_name, to_entity_list):
        return cls._link(from_uuId, to_entity_name, to_entity_list, 'update')

    @classmethod
    def _delete_link(cls, from_uuId, to_entity_name, to_entity_list, list_name=None):
        return cls._link(from_uuId, to_entity_name, to_entity_list, 'delete', list_name=list_name)

    @classmethod
    def _link(cls, from_uuId, to_entity_name, to_entity_list, operation, list_name=None):
        """
        `from_uuId`: Source entity uuId

        `to_entity_name`: Destination entity name (e.g. 'staff')

        `to_entity_list`: List of uuIds (and optional data) we are linking to

        `operation`: `add`, `update`, `delete`

        `list_name`: Some list names differ from the standard pattern, so we
         pass in the name we would like to use instead.
        """
        to_key = list_name or to_entity_name.lower() + 'List'
        payload = {
            'uuId': from_uuId,
            to_key: to_entity_list
        }
        url = '/api/{}/link/{}/{}'.format(cls._path, to_entity_name, operation)
        api.post(url, payload=payload)
        return True


class AccessPolicyLinker(BaseLinker):
    """Subclass can link to Access Policies"""
    def link_access_policy(self, access_policy):
        self._add_link(self['uuId'], 'access_policy', [access_policy], list_name='accessPolicyList')

    def unlink_access_policy(self, access_policy):
        self._delete_link(self['uuId'], 'access_policy', [access_policy], list_name='accessPolicyList')


class CompanyLinker(BaseLinker):
    """Subclass can link to Companies"""
    def link_company(self, company):
        self._add_link(self['uuId'], 'company', [company])

    def unlink_company(self, company):
        self._delete_link(self['uuId'], 'company', [company])


class ContactLinker(BaseLinker):
    """Subclass can link to Contacts"""
    def link_contact(self, contact):
        self._add_link(self['uuId'], 'contact', [contact])

    def unlink_contact(self, contact):
        self._delete_link(self['uuId'], 'contact', [contact])


class CustomerLinker(BaseLinker):
    """Subclass can link to Customers"""
    def link_customer(self, customer):
        self._add_link(self['uuId'], 'customer', [customer])

    def unlink_customer(self, customer):
        self._delete_link(self['uuId'], 'customer', [customer])


class DepartmentLinker(BaseLinker):
    """Subclass can link to Departments"""
    def link_department(self, department):
        self._add_link(self['uuId'], 'department', [department])

    def unlink_department(self, department):
        self._delete_link(self['uuId'], 'department', [department])


class FileLinker(BaseLinker):
    """Subclass can link to Files"""
    def link_file(self, file):
        self._add_link(self['uuId'], 'file', [file], list_name='storageFileList')

    def unlink_file(self, file):
        self._delete_link(self['uuId'], 'file', [file], list_name='storageFileList')


class FolderLinker(BaseLinker):
    """Subclass can link to Folders"""
    def link_folder(self, folder):
        self._add_link(self['uuId'], 'folder', [folder], list_name='folders')

    def unlink_folder(self, folder):
        self._delete_link(self['uuId'], 'folder', [folder], list_name='folders')


class LocationLinker(BaseLinker):
    """Subclass can link to Locations"""
    def link_location(self, location):
        self._add_link(self['uuId'], 'location', [location])

    def unlink_location(self, location):
        self._delete_link(self['uuId'], 'location', [location])


class PermissionLinker(BaseLinker):
    """Subclass can link to Permissions"""
    def link_permission(self, permission):
        return self._add_link(self['uuId'], 'permission', [permission])

    def unlink_permission(self, permission):
        return self._delete_link(self['uuId'], 'permission', [permission])


class ProjectLinker(BaseLinker):
    """Subclass can link to Projects"""
    def link_project(self, project):
        self._add_link(self['uuId'], 'project', [project])

    def unlink_project(self, project):
        self._delete_link(self['uuId'], 'project', [project])


class RebateLinker(BaseLinker):
    """Subclass can link to Rebates"""
    def link_rebate(self, rebate):
        self._add_link(self['uuId'], 'rebate', [rebate])

    def unlink_rebate(self, rebate):
        self._delete_link(self['uuId'], 'rebate', [rebate])


class ResourceLinker(BaseLinker):
    """Subclass can link to Resources"""
    def link_resource(self, resource):
        self._add_link(self['uuId'], 'resource', [resource])

    def relink_resource(self, resource):
        self._update_link(self['uuId'], 'resource', [resource])

    def unlink_resource(self, resource):
        self._delete_link(self['uuId'], 'resource', [resource])


class SkillLinker(BaseLinker):
    """Subclass can link to Skills"""
    def link_skill(self, skill):
        self._add_link(self['uuId'], 'skill', [skill])

    def relink_skill(self, skill):
        self._update_link(self['uuId'], 'skill', [skill])

    def unlink_skill(self, skill):
        self._delete_link(self['uuId'], 'skill', [skill])


class StaffLinker(BaseLinker):
    """Subclass can link to Staff"""
    def link_staff(self, staff):
        self._add_link(self['uuId'], 'staff', [staff])

    def relink_staff(self, staff):
        self._update_link(self['uuId'], 'staff', [staff])

    def unlink_staff(self, staff):
        self._delete_link(self['uuId'], 'staff', [staff])


class StageLinker(BaseLinker):
    """Subclass can link to Stages"""
    def link_stage(self, stage):
        self._add_link(self['uuId'], 'stage', stage, list_name='stage')

    def unlink_stage(self, stage):
        self._delete_link(self['uuId'], 'stage', stage, list_name='stage')


class StageListLinker(BaseLinker):
    """Subclass can link to Stage List"""
    def link_stage_list(self, stage):
        self._add_link(self['uuId'], 'stage_list', [stage], list_name='stageList')

    def unlink_stage_list(self, stage):
        self._delete_link(self['uuId'], 'stage_list', [stage], list_name='stageList')
